'''
    This script tool allocates origins having supply to destinations having demand 
such that the supply is consumed, the demands are met, and the overall
transportation cost is minimized. This tool can be used to solve common allocation
problems having capacity constraints such as allocating stores to warehouses or
allocating students to schools. 
    The tool solves the classic transportation problem by first generating a travel 
cost matrix between the given origins and destinations using the OD Cost matrix
solver provided by the network analyst extension. The tool then formulates and
solves a linear programming problem (LPP) using an open source LPP solver, COIN, to
determine the optimum allocation. 
    This tool requires  an external python site-package called Pulp. Download the
latest version of Pulp from http://pypi.python.org/pypi/PuLP and install Pulp by 
extracting the zip file to a folder and running "setup.py install" from that folder.
The tool works only with ArcGIS version 10. The tool requires ArcGIS Network Analyst
extension and can be run with an ArcView license. 

Please send Questions/Comments/Suggestions to  Deelesh Mandloi, dmandloi@esri.com
Last Modified November 16, 2011
'''

def Trace():
    '''return the filename, line number and error message in case of exception'''
    #Get the traceback object
    tb = sys.exc_info()[2]
    #Get the script name and line number from the trace back object
    tbinfo = traceback.format_tb(tb)[0]
    line = tbinfo.split(", ")[1]
    #filename = os.path.join(sys.path[0],"OptimumAllocation.py")
    filename  = __file__
    #Get the python error
    pyError = traceback.format_exc().splitlines()[-1]
    return line,filename,pyError
      

def GetNAClasses(nalayer):
    '''Get all the sublayers from input NALayer. This will not work if using a NALayer as lyr file'''    
    NAClasses = {} #Dict that stores NAClassName:NAClassLayerName key value pair
    desc = arcpy.Describe(nalayer)
    sublayers = desc.children
    for sublayer in sublayers:
        NAClasses[sublayer.name.lower()] = sublayer.namestring
    return NAClasses

def deleteNALayer(nalayer):
    '''Deletes the inmemory NA layer if it exists'''
    if arcpy.Exists(nalayer):
        arcpy.Delete_management(nalayer)
        
def PrintAndWriteMessages(msg,severity=0):
    '''prints a GP message and writes the same message to a log file'''
    if severity == 0:
        arcpy.AddMessage(msg)
    elif severity == 1:
        arcpy.AddWarning(msg)
    elif severity == 2:
        arcpy.AddError(msg)
    logfile.write(msg + "\n")

def GetFeaturesWithPositiveValues(features, field):
    '''Returns a feature layer that contains only the features with positive values for the field.
    Adds warning message with a list of feature id's if features have negative or null values.'''
    #Create a feature layer from input features if it is not already a feature layer.    
    dataType = arcpy.Describe(features).dataType.lower()
    if dataType == "featurelayer":
        featureLayer = features
    else:
        featureLayer = "LayerWithFeaturesToLoad"
        arcpy.MakeFeatureLayer_management(features, featureLayer)
    #get a count for total features
    featureCount = long(arcpy.GetCount_management(featureLayer).getOutput(0))
    if not featureCount:
        return featureLayer
    #Find features with negative or null values
    whereClause = field + " <= 0 OR " + field + " IS NULL"
    #ignoreFeatureLayer = "FeaturesToIgnore"
    arcpy.SelectLayerByAttribute_management(featureLayer, "NEW_SELECTION", whereClause)
    ignoreFeatures = arcpy.Describe(featureLayer).FIDSet.replace(";", ",")
    if ignoreFeatures:
        PrintAndWriteMessages("\tThe features with following OIDs from %s are not used as they don't have positive values for %s field." % (features, field),1)
        PrintAndWriteMessages("\t%s" % ignoreFeatures,1)
    arcpy.SelectLayerByAttribute_management(featureLayer, "SWITCH_SELECTION")
    return featureLayer
        

class LPSolverError(Exception):
    '''Exception returned by the LPPSolver'''
    pass
    
#import the modules
import sys
import os
import traceback
import time
import arcpy

#overwrite existing outputs
arcpy.env.overwriteOutput = True

#Create a log file to contain messages in the same location as the script
logfileName = os.path.splitext(os.path.basename(__file__))[0] + "_log.txt"
logfile = open(os.path.join(sys.path[0],logfileName),"w",0)
logfile.write("Start time: " + time.asctime() + "\n")

try:     
    #Check out the network extension. Quit if not found
    if arcpy.CheckExtension("network").lower() == "available":
        arcpy.CheckOutExtension("network")
    else:
        sys.exit(1)
    # Import PuLP module
    import pulp
    #Create a progressor
    arcpy.SetProgressor("step","Computing Optimum Allocation",0,15,1)
    
    #Get the supply and demand field names and origin and dest fc's
    origin_fc = arcpy.GetParameterAsText(0)
    supplyfield = arcpy.GetParameterAsText(1)
    ratefield = arcpy.GetParameterAsText(2)
    dest_fc = arcpy.GetParameterAsText(3)
    demandfield = arcpy.GetParameterAsText(4)
    
    #get the required inputs from the tool dialog to create the OD layer
    nds = arcpy.GetParameterAsText(5)
    impedance = arcpy.GetParameterAsText(6)
    impedance_cutoff = arcpy.GetParameterAsText(7)
    netlayer = arcpy.GetParameterAsText(8)
    usehir = arcpy.GetParameterAsText(9)
    uturn_policy = arcpy.GetParameterAsText(10)
    restrictions = arcpy.GetParameterAsText(11)
    barrier_fc = arcpy.GetParameterAsText(12)
    search_tolerance = arcpy.GetParameterAsText(13)
    reportInfeasibile = arcpy.GetParameter(14)
    msg = "Creating a new OD Cost Matrix Layer"
    PrintAndWriteMessages(msg,0)
    arcpy.SetProgressorLabel(msg)
    
    # describe the nds to get value of accumulators parameter
    # we wish to accumulate all of the cost attributes
    desc = arcpy.Describe(nds)
    attrUnits = {}
    accumulators = ""
    attributes = desc.attributes
    for attribute in attributes:
        if attribute.UsageType == "Cost":
            accumulators += attribute.Name + ";"
            attrUnits[attribute.Name] = attribute.Units
    accumulators = accumulators.rstrip(";")
    
    #get and set the optional parameters to create the OD layer
    if impedance_cutoff == "":
        impedance_cutoff == "#"
    if usehir.lower() == 'true':
        hierarchy = "USE_HIERARCHY"
    else:
        hierarchy = "NO_HIERARCHY"
        
    #Create a new OD Costmatrix layer
    arcpy.MakeODCostMatrixLayer_na(nds,netlayer,impedance,impedance_cutoff,"#",accumulators,uturn_policy,restrictions,hierarchy,"#","STRAIGHT_LINES")
    #Add various fields on the origins, destinations and lines sublayers in OD layer.
    #Add the supply and demand fields
    arcpy.AddFieldToAnalysisLayer_na(netlayer,"Origins","Supply","Double")
    arcpy.AddFieldToAnalysisLayer_na(netlayer,"Destinations","Demand","Double")
    arcpy.AddFieldToAnalysisLayer_na(netlayer,"Origins","CostAtOrigin","Double")
    arcpy.AddFieldToAnalysisLayer_na(netlayer,"Lines","Total_CostAtOrigin","Double")
    #add the extra supply and extra demand fields and assign them default values of 0
    arcpy.AddFieldToAnalysisLayer_na(netlayer, "Origins", "ExtraSupply", "Double")
    arcpy.AddFieldToAnalysisLayer_na(netlayer, "Destinations", "ExtraDemand", "Double")
    # Add the field to store the optimal allocation of units after TP is solved
    arcpy.AddFieldToAnalysisLayer_na(netlayer, "Lines", "Allocate_Units","Double")
    # Add the field to store if the origin or destination was included in the OD matrix
    #This is useful to see which origins were not used due to cutoff values
    arcpy.AddFieldToAnalysisLayer_na(netlayer,"Origins", "OriginUsed","Short")
    arcpy.AddFieldToAnalysisLayer_na(netlayer,"Destinations", "DestinationUsed","Short")

    arcpy.SetProgressorPosition()
    
    #Load the restriction barriers into appropriate na classes based on input shape type.
    #Do this before loading origins and destinations so as to avoid locating them on restricted elements
    if not barrier_fc in ("","#"):
        msg = "Loading Restriction Barriers"
        PrintAndWriteMessages(msg,0)
        arcpy.SetProgressorLabel(msg) 
        barrierShapeType = arcpy.Describe(barrier_fc).shapeType.lower()
        #load the barriers using the default field mappings        
        if barrierShapeType == 'point':
            arcpy.AddLocations_na(netlayer,"Point Barriers",barrier_fc,"#",search_tolerance)
        elif barrierShapeType == 'polyline':
            arcpy.AddLocations_na(netlayer,"Line Barriers",barrier_fc,"#",search_tolerance)
        elif barrierShapeType == 'polygon':
            arcpy.AddLocations_na(netlayer,"Polygon Barriers",barrier_fc,"#",search_tolerance)
    
    arcpy.SetProgressorPosition()
    
    #Load Origins and Destinations in NALayer along with values from Supply and demand fields
    msg = "Loading Origins"
    PrintAndWriteMessages(msg,0)
    arcpy.SetProgressorLabel(msg)
    #load only those origins that have a positive non null supply value
    originsToLoad = GetFeaturesWithPositiveValues(origin_fc, supplyfield)
    #Field map is "FieldSrc FieldTarget #; FieldSrc FieldTarget #"
    #if no ratefield is specified use a constant rate value of 1
    useRate = True
    if ratefield == "":
        useRate = False    
    if useRate:
        Origin_FieldMap = "Supply " + supplyfield + " #" + ";" + "CostAtOrigin " + ratefield + " #" + ";" + " Name Name #; CurbApproach # 0" 
    else:
        Origin_FieldMap = "Supply " + supplyfield + " #" + ";" + "CostAtOrigin # 1" + ";" + " Name Name #; CurbApproach # 0" 
    arcpy.AddLocations_na(netlayer,"Origins",originsToLoad,Origin_FieldMap,search_tolerance,"","","","","","",
                          "EXCLUDE")
    arcpy.SelectLayerByAttribute_management(originsToLoad, "CLEAR_SELECTION")
    arcpy.SetProgressorPosition()
    
    msg = "Loading Destinations"
    PrintAndWriteMessages(msg,0)
    arcpy.SetProgressorLabel(msg)
    #load only those Destinations that have a positive and non null demand value.
    destinationsToLoad = GetFeaturesWithPositiveValues(dest_fc, demandfield)
    Dest_FieldMap = "Demand " + demandfield + " #" + "; CurbApproach # 0; Name Name #" 
    arcpy.AddLocations_na(netlayer,"Destinations",destinationsToLoad,Dest_FieldMap,search_tolerance,"","","","","","",
                          "EXCLUDE")
    arcpy.SelectLayerByAttribute_management(destinationsToLoad, "CLEAR_SELECTION")
    arcpy.SetProgressorPosition()
    
    #Solve the od layer. Error out if there is invalid location by using HALT option of Solve  
    msg = "Solving the OD Cost Matrix"
    PrintAndWriteMessages(msg,0)
    arcpy.SetProgressorLabel(msg) 
    arcpy.Solve_na(netlayer,"HALT")
    arcpy.SetProgressorPosition()
    
    msg = "Formulating the linear programming model"
    PrintAndWriteMessages(msg,0)
    msg = "\tStep1: Getting all the costs...."
    PrintAndWriteMessages(msg,0)
    arcpy.SetProgressorLabel(msg.lstrip('\t'))
    
    # Get the OD matrix sub layers from entire analysis layer.
    odnaclasses = GetNAClasses(netlayer)
    odmatrix = odnaclasses["odlines"]
    originlayer = odnaclasses["origins"]
    destlayer = odnaclasses["destinations"]


    #Calculate the new Total_CostAtOrigin by getting the transport rate from origins and multiplyng by Total_Impedance
    #For this transfer the CostAtOrigin from Origins to Lines
    if useRate:
        #Add the join
        arcpy.AddJoin_management(odmatrix, "OriginID", originlayer, "ObjectID", "KEEP_ALL")
        #Calculate the field
        field_expression = "!ODLines.Total_" + impedance + "! * !Origins.CostAtOrigin!"
        arcpy.CalculateField_management(odmatrix, "ODLines.Total_CostAtOrigin", field_expression, "PYTHON_9.3", "#")
        #Remove the join
        arcpy.RemoveJoin_management(odmatrix, "Origins")
        costfield = "Total_CostAtOrigin"
    else:
        costfield = "Total_" + impedance
    #Get the cursor for all records in the matrix.
    odmatrixcur = arcpy.SearchCursor(odmatrix)

    #Create an empty dictionary to contain all the decision variables. i.e all OD pairs and associated transport cost
    costdict = {}
    for row in odmatrixcur:
        # each decision variable is "x_OriginID_DestinationID"
        var = "x_" + str(row.getValue("OriginID")) + "_" + str(row.getValue("DestinationID"))    
        cost = row.getValue(costfield)
        #fill the value for each element in the matrix
        costdict[var] = cost
    del odmatrixcur, row
    arcpy.SetProgressorPosition()
    
    msg = "\tStep2: Creating decision variables...."
    PrintAndWriteMessages(msg,0)
    arcpy.SetProgressorLabel(msg.lstrip("\t"))
    
    # Start a new LP problem
    prob = pulp.LpProblem("test", pulp.LpMinimize)

    # Get all the decision variables as keys from the dictionary
    odpairs = costdict.keys()

    #Create the LP decision variables and store them in a list. Also get the cost for that variable.
    decvars = []
    odcost = []
    for x in odpairs:
        #Lpvariable creates non negative variable whose name is same a dict key
        decvars.append(pulp.LpVariable(x, 0))
        odcost.append(costdict[x])

    arcpy.SetProgressorPosition()
    
    # Create the objective function using lpDot to get Dot product of two lists
    msg = "\tStep3: Building objective function with costs and decision variables...."
    PrintAndWriteMessages(msg,0)
    arcpy.SetProgressorLabel(msg.lstrip('\t'))
    obj = pulp.lpDot(odcost,decvars)
    prob += obj

    arcpy.SetProgressorPosition()
    
    msg = "\tStep4: Getting all the supplies...."
    PrintAndWriteMessages(msg,0)
    arcpy.SetProgressorLabel(msg.lstrip('\t'))
    
    #get only unique oids from all oids
    uoids = []
    for x in odpairs:
        uoid = x.split("_")[1]
        if not uoid in uoids:        
            uoids.append(uoid)
    uoids.sort()    

    #get all the unique destination ids from OD pairs
    udids = []
    for x in odpairs:    
        udid = x.split("_")[2]
        if not udid in udids:
            udids.append(x.split("_")[2])
    udids.sort()

    #get the supply for all the origins by using the origins layer in OD matrix layer
    origincur = arcpy.UpdateCursor(originlayer)
    
    #Store the supply in a dict with key value as ObjectID which is Origin ID in matrix layer
    supplydict = {}
    unusedOrigins = []
    for row in origincur:
        cost = row.getValue("Supply")
        oid = str(row.getValue("ObjectID"))
        #if cutoff is specified, then all origins may not be in OD matrix
        #So remove those origins and update OriginUsed field        
        if oid in uoids:
            supplydict[oid] = cost
            row.setValue("OriginUsed",1)
            #Set the intial extra supply value to 0. This will be updated after the solution is found.
            row.setValue("ExtraSupply", 0)
        else:
            unusedOrigins.append(row.getValue("Name"))
            row.setValue("OriginUsed",0)
            row.setValue("ExtraSupply", row.getValue("Supply"))
        origincur.updateRow(row)
    del origincur, row
    if unusedOrigins:
        PrintAndWriteMessages("\tNo routes were found from the following Origins based on the specified network analysis options.", 1)
        PrintAndWriteMessages("\t" + ", ".join(unusedOrigins),1)
    arcpy.SetProgressorPosition()
    
    msg = "\tStep5: Getting all the demands...."
    PrintAndWriteMessages(msg,0)
    arcpy.SetProgressorLabel(msg.lstrip('\t'))
    #get the demand for all the destinations by using the Destinations layer in OD matrix solver
    destincur = arcpy.UpdateCursor(destlayer)
    #Store the demand in a dict with key value as ObjectID which is destination ID in matrix layer
    demanddict = {}
    unreachedDestinations = []    
    for row in destincur:
        cost = row.getValue("Demand")
        oid = str(row.getValue("ObjectID"))
        #if cutoff is specified, then all destinations may not be in OD matrix
        #So remove those destinations and update DestinationUsed field
        if oid in udids:
            demanddict[oid] = cost
            row.setValue("DestinationUsed",1)
            #Set the intial extra demand value to 0. This will be updated after the solution is found.
            row.setValue("ExtraDemand", 0)
        else:
            unreachedDestinations.append(row.getValue("Name"))
            row.setValue("DestinationUsed",0)
            row.setValue("ExtraDemand", row.getValue("Demand"))
        destincur.updateRow(row)
    del destincur, row
    if unreachedDestinations:
        PrintAndWriteMessages("\tNo routes were found to the following Destinations due to presence of an impedance cutoff or other network analysis options.", 1)
        PrintAndWriteMessages("\t" + ", ".join(unreachedDestinations),1)
    arcpy.SetProgressorPosition()
    
    msg = "\tStep6: Formualting supply constraints...."
    PrintAndWriteMessages(msg,0)
    arcpy.SetProgressorLabel(msg.lstrip('\t'))
    
    #Check if Supply is greater than demand or viceversa
    totalsupply = sum(supplydict.values())
    totaldemand = sum(demanddict.values())
    if totalsupply > totaldemand:
        #Excess Supply Case
        probcase = "extrasupply"
        extrasupply = totalsupply - totaldemand
        #Create an extra column vector of fictious demand destination
        #This column will have extra decision variables equal to number to origins i.e f1 to f4 if we have 4 origins
        #So if 5 origins then 5 slack variables f_originid
        slackvars = []
        for i in uoids:
            slackDecVar = pulp.LpVariable("f_o_"+str(i),0)
            decvars.append(slackDecVar)
            slackvars.append(slackDecVar)
        #Add an extra demand contriant which is sum of all slack varibales
        slackcoeff = [1]
        slackcoeff = slackcoeff * len(uoids)
        prob += pulp.lpDot(slackcoeff,slackvars) == extrasupply
    elif totalsupply < totaldemand:
        probcase = "extrademand"
        extrademand = totaldemand - totalsupply
        slackvars = []
        for j in udids:
            slackDecVar = pulp.LpVariable("f_d_"+str(j),0)
            decvars.append(slackDecVar)
            slackvars.append(slackDecVar)
        #Add an extra supply contriant which is sum of all slack varibales
        slackcoeff = [1]
        slackcoeff = slackcoeff * len(udids)
        prob += pulp.lpDot(slackcoeff,slackvars) == extrademand
    else:
        probcase = "balanced"
        
    #Formulate the supply (row) constraints
    for x in uoids:
        # create two empty lists to store a row in the matrix
        rowdecvars = []
        rowcoeff = [1]
        # Get only those decvars where oid is matching
        for decvar in decvars:
            decVarName = str(decvar).split("_")
            if probcase == "extrasupply" and (decVarName[1] == x): 
                rowdecvars.append(decvar)
            elif probcase == "extrademand" and (decVarName[1] == x): 
                if decVarName[0] != 'f':            
                    rowdecvars.append(decvar)
            elif probcase == "balanced":
                if decVarName[1] == x:
                    rowdecvars.append(decvar)
        if probcase == "extrasupply":
            rowdecvars.append(pulp.LpVariable('f_o_'+ x,0))
        rowcoeff = rowcoeff * len(rowdecvars)
        prob += pulp.lpDot(rowcoeff,rowdecvars) == supplydict[x]
    
    arcpy.SetProgressorPosition()
    
    msg = "\tStep7: Formulating demand constraints...."
    PrintAndWriteMessages(msg,0)
    arcpy.SetProgressorLabel(msg.lstrip('\t'))
    
    #Formulate the demand (column) constraints
    for x in udids:
        # create two empty lists to store a column in the matrix
        coldecvars = []
        colcoeff = [1]
        # Get only those decvars where did is matching
        for decvar in decvars:
            decVarName = str(decvar).split("_")
            if probcase == "extrademand" and (decVarName[2] == x):
                coldecvars.append(decvar)
            elif probcase == "extrasupply" and (decVarName[2] == x):
                if decVarName[0] != 'f':
                    coldecvars.append(decvar)
            elif probcase == "balanced":
                if decVarName[2] == x:
                    coldecvars.append(decvar)
                
        #create the unity coeffcients list for each col element
        colcoeff = colcoeff*len(coldecvars)
        prob += pulp.lpDot(colcoeff,coldecvars) == demanddict[x]

    arcpy.SetProgressorPosition()
    
    #Write the problem as an LP file. File created in same folder as the script
    try:
        prob.writeLP(os.path.join(sys.path[0],"OptimumAllocation.lp"))
    except pulp.PulpError as ex:
        PrintAndWriteMessages("Failed to write the problem as a LP file",1)
        
    
    #Solve the problem using the default solver. If the solve fails try to resolve
    #using the COINMP DLL solver.
    msg = "Solving the linear programming model"
    PrintAndWriteMessages(msg,0)
    arcpy.SetProgressorLabel(msg)
    try:
        prob.solve()
    except pulp.PulpSolverError:
        PrintAndWriteMessages("Failed to solve the LP using {0} solver".format(pulp.LpSolverDefault.__class__),1)
        try:
            prob.setSolver(pulp.COINMP_DLL())
            PrintAndWriteMessages("Resolving the LP using {0} solver".format(prob.solver.__class__))
            prob.resolve()
        except:
            raise LPSolverError("Failed to solve the LP.")
    except:
        raise
    
    # Print the status of the solved LP
    solverstatus = str(pulp.LpStatus[prob.status]).lower()    
    msg = "\tLP solver status: " + solverstatus
    PrintAndWriteMessages(msg)
    
    if solverstatus == "infeasible":
        if reportInfeasibile:    
            msg = "\tThe solution is not optimal. It may not be correct."
            PrintAndWriteMessages(msg,1)
        else:
            raise LPSolverError("The LP solver failed to find an optimal solution.")
    
    arcpy.SetProgressorPosition()
    msg = "Writing the output from the LP solver"
    PrintAndWriteMessages(msg,0)
    arcpy.SetProgressorLabel(msg)
            
    # Write the optimal Allocation to a new field in matrix 
    # Create an update cursor
    allocatecur = arcpy.UpdateCursor(odmatrix)
    varValues = {} #Store varName: varValue    
    for v in prob.variables():
        if str(v.varValue).lower() != "none":
            varValues[str(v.name)] = v.varValue
    for row in allocatecur:
        rowodpairid = "x_" + str(row.getValue("OriginID")) + "_" + str(row.getValue("DestinationID"))
        if rowodpairid in varValues:
            row.setValue("Allocate_Units", varValues[rowodpairid])
            allocatecur.updateRow(row)
    del allocatecur,row
    
    #Delete OD lines from the matrix where there is no allocation
    arcpy.SelectLayerByAttribute_management(odmatrix,"NEW_SELECTION", 'Allocate_Units = 0')
    arcpy.DeleteRows_management(odmatrix)
    
    #Print the value of the objective and set it as output value
    #use impedance units as cost units
    costUnits = attrUnits[impedance]
    msg = "\tTotal cost = " + str(pulp.value(prob.objective)) + " " + costUnits 
    PrintAndWriteMessages(msg,0)
    arcpy.SetParameterAsText(16,pulp.value(prob.objective))
    
    #if unbalanced, then write the appropriate excess supply or demand
    if probcase == "extrasupply":
        #Extra supply. So add a column of slack variable values to origins
        extrasupplycur = arcpy.UpdateCursor(originlayer)
        for row in extrasupplycur:
            roworiginid = 'f_o_' + str(row.getValue("ObjectID"))
            if roworiginid in varValues:
                row.setValue("ExtraSupply",varValues[roworiginid])
                extrasupplycur.updateRow(row)
        del extrasupplycur,row
    
    if probcase == "extrademand":
        #Extra demand. So add a column of slack variable values to destinations
        extrademandcur = arcpy.UpdateCursor(destlayer)
        for row in extrademandcur:
            rowdestid = 'f_d_'+ str(row.getValue("ObjectID"))
            if rowdestid in varValues:
                row.setValue("ExtraDemand",varValues[rowdestid])
                extrademandcur.updateRow(row)
        del extrademandcur,row
            
    arcpy.SetProgressorPosition()    
    #Add the output OD layer to ArcMap TOC
    arcpy.SetParameterAsText(15,netlayer)

except LPSolverError as ex:
    #Handle LPSolver errors
    PrintAndWriteMessages(ex.message,2)
    if 'netlayer' in dir():
        deleteNALayer(netlayer)
except arcpy.ExecuteError:
    #Handle GP specific errors
    line, filename, err = Trace()
    msg = "Geoprocessing error on " + line + " of " + filename + " :" + "\n" + arcpy.GetMessages(1) + arcpy.GetMessages(2)
    PrintAndWriteMessages(msg,2) 
    if 'netlayer' in dir():
        deleteNALayer(netlayer)
except ImportError:
    #Handle pulp import error
    msg = "Pulp module not installed."
    PrintAndWriteMessages(msg,2)
    msg = "Download and install Pulp from http://pypi.python.org/pypi/PuLP"
    PrintAndWriteMessages(msg,1)
except SystemExit:
    #Handle system exit when sys.exit() is called
    msg = "Network Analyst license not found"
    PrintAndWriteMessages(msg,2)
except:
    #Handle Python errors
    line, filename, err = Trace()
    msg = "Python error on " + line + " of " + filename + "\n" + err 
    PrintAndWriteMessages(msg,2)
    if 'netlayer' in dir():
        deleteNALayer(netlayer)
finally:
    #Close the log file with a closing time stamp
    logfile.write("End Time: " + time.asctime())
    logfile.close()
    
    
    
